/**
 * hede.c
 *
 * Çok basit proc modülü.
 * Proces bir ara belleğe bilgi yazar veya yazılan bilgiyi okur.
 *
 * Kaynak:
 * The Linux Kernel Module Programming Guide, procfs3.c'den uyarlanmıştır.
 *
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>

#define FILE_NAME "aselsan" //  /proc/aselsan
#define MAX_SIZE   1024

static char kernel_buf[MAX_SIZE];

/*
 *  kernel_buf[x x x x x x - - - - - ] 
 *                         ^          ^
 *            kernel_buf_len    MAX_SIZE 
 *
 */
static unsigned long kernel_buf_len= 0;

/**
 * /proc/test'den okuma yapıldığı zanan bu fonksiyon çağrılır.
 *
 * Kernel bölgesinde kernel_buf[] içinde bulunan kernel_buf_len adet karakter 
 * buffer[] aracılığı ile user space tarafına kopyalanır.
 *
 * Kernel, ilk çağrıda offset'i 0 gönderir.
 *
 * Kernel, read işlemini eof bulana kadar devam ettirir.
 * İlk read çağrısında buffer[]'a yazılan miktar döner.
 * Sonraki read'de ise 0 dönmelidir ki kernel eof geldiğini anlasın.
 *
 * ilk read okuması aşağıdaki gibi işler
 *    finished=0
 *    copy data to user space
 *    finished= 1
 *    kopyalanan bayt sayısı döner.
 *
 * Sonraki read okuması aşağıdaki gibi işler
 *    finished artık 1'dir.
 *    finished= 0; yapılır ki sonraki okumada tekrar aynı veri gönderilsin.
 *    0 sayısı döner, bu da çekirdeğe eof geldiğini söyler.
 *
 * Eğer finished kullanılmamış olsaydı, sonsuz döngü oluşurdu.
 *
 */
static ssize_t test_read(struct file *filp,    
                         char        *buffer,
                         size_t       len,    
                         loff_t      *offset
		        ){

    static int finished = 0;

    if ( finished ) {
       printk(KERN_INFO "%s(): end\n", __FUNCTION__);
       finished = 0;
       return 0;  // eof
       }
    
    finished= 1;
        
    /* Kernel bellek bölgesinden user space tarafına veri kopyala.
     *
     */
    if ( copy_to_user(buffer, kernel_buf, kernel_buf_len) ) {
       return -EFAULT;
       }

    printk(KERN_INFO "%s(): %lu bayt okundu.\n", __FUNCTION__, kernel_buf_len);

    return kernel_buf_len;    // okunan eleman sayısı döner.
}

/**
 * /proc/test'ya yazım yapıldığı zaman bu fonksiyon çağrılır.
 *
 * user space tarafındaki buffer[] adresinde bulunan len adet 
 * bilgi kernel space tarafındaki buf[] alanına kopyalanacaktır.
 *
 */
static ssize_t test_write(struct file *file, 
		          const char  *buffer, 
			  size_t       len, 
			  loff_t      *off
			 ){

    if ( len > MAX_SIZE ){
       kernel_buf_len= MAX_SIZE;
       }
       else{
       kernel_buf_len= len;
       }
    
    /* User space bellek bölgesinden kernel bellek bölgesine veri kopyala.
     *
     */
    if ( copy_from_user(kernel_buf, buffer, kernel_buf_len) ) {
       return -EFAULT;
       }

    printk(KERN_INFO "%s(): %lu bayt yazıldı.\n", __FUNCTION__, kernel_buf_len);
    
    return kernel_buf_len;
}

/**
 * File operatrions.
 *
 */
static const struct file_operations fops= {
    .owner  = THIS_MODULE,
    .read   = test_read,
    .write  = test_write
    };

/**
 * Modül yüklendikten hemen sonra burası çağrılır.
 *
 */
static int __init example_init(void){

   if ( proc_create(FILE_NAME, 0666, NULL, &fops) == NULL) {
      return -ENOMEM;
      }

   printk(KERN_INFO "%s kuruldu\n", FILE_NAME);

   return 0;
}

/**
 * Modül sonlanırken burası çağrılır.
 *
 */
static void __exit example_exit(void){
    remove_proc_entry(FILE_NAME, NULL);
    printk(KERN_INFO "%s yok edildi\n", FILE_NAME);
}

/**
 * init ve exit fonksiyon isimleri standard değil.
 * Çekirdeğe isimleri ayrıca söyle.
 *
 */
module_init(example_init);
module_exit(example_exit);

/**
 * Metin bilgileri.
 *
 */
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Naazim Koch http://UCanLinux.Com");
MODULE_DESCRIPTION("very simple proc module");
MODULE_VERSION("0.0001");

